/* ===========================================================================
Copyright (C) 2002-2005  3Dlabs Inc. Ltd.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

    Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

    Redistributions in binary form must reproduce the above
    copyright notice, this list of conditions and the following
    disclaimer in the documentation and/or other materials provided
    with the distribution.

    Neither the name of 3Dlabs Inc. Ltd. nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
=========================================================================== */

#include "ogl2brick.h"

//
// Global handles for the currently active program object, with its two shader objects
//
GLuint ProgramObject = 0;
GLuint VertexShaderObject = 0;
GLuint FragmentShaderObject = 0;

GLint Video = 0;
GLint RenderBrickPattern = 0;
GLint SwitchShaders = 0;
GLint LightPosition = 0;
GLint TexLoc = 0;

//
// Get the location of a uniform variable
//
static
GLint getUniLoc(GLuint program, const GLchar *name)
{
    GLint loc;

    loc = glGetUniformLocation(program, name);

    if (loc == -1)
        printf("No such uniform named \"%s\"\n", name);

    printOpenGLError();  // Check for OpenGL errors
    return loc;
}

//
// Print out the information log for a shader object
//
static
void printShaderInfoLog(GLuint shader)
{
    int infologLength = 0;
    int charsWritten  = 0;
    GLchar *infoLog;

    printOpenGLError();  // Check for OpenGL errors

    glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &infologLength);

    printOpenGLError();  // Check for OpenGL errors

    if (infologLength > 0)
    {
        infoLog = (GLchar *)malloc(infologLength);
        if (infoLog == NULL)
        {
            printf("ERROR: Could not allocate InfoLog buffer\n");
            exit(1);
        }
        glGetShaderInfoLog(shader, infologLength, &charsWritten, infoLog);
        printf("Shader InfoLog:\n%s\n\n", infoLog);
        free(infoLog);
    }
    printOpenGLError();  // Check for OpenGL errors
}


//
// Print out the information log for a program object
//
static
void printProgramInfoLog(GLuint program)
{
    int infologLength = 0;
    int charsWritten  = 0;
    GLchar *infoLog;

    printOpenGLError();  // Check for OpenGL errors

    glGetProgramiv(program, GL_INFO_LOG_LENGTH, &infologLength);

    printOpenGLError();  // Check for OpenGL errors

    if (infologLength > 0)
    {
        infoLog = (GLchar *)malloc(infologLength);
        if (infoLog == NULL)
        {
            printf("ERROR: Could not allocate InfoLog buffer\n");
            exit(1);
        }
        glGetProgramInfoLog(program, infologLength, &charsWritten, infoLog);
        printf("Program InfoLog:\n%s\n\n", infoLog);
        free(infoLog);
    }
    printOpenGLError();  // Check for OpenGL errors
}


static
int shaderSize(char *fileName, EShaderType shaderType)
{
    //
    // Returns the size in bytes of the shader fileName.
    // If an error occurred, it returns -1.
    //
    // File name convention:
    //
    // <fileName>.vert
    // <fileName>.frag
    //
    int fd;
    char name[100];
    int count = -1;

    strcpy(name, fileName);

    switch (shaderType)
    {
        case EVertexShader:
            strcat(name, ".vert");
            break;
        case EFragmentShader:
            strcat(name, ".frag");
            break;
        default:
            printf("ERROR: unknown shader file type\n");
            exit(1);
            break;
    }

    //
    // Open the file, seek to the end to find its length
    //
#ifdef WIN32 /*[*/
    fd = _open(name, _O_RDONLY);
    if (fd != -1)
    {
        count = _lseek(fd, 0, SEEK_END) + 1;
        _close(fd);
    }
#else /*][*/
    fd = open(name, O_RDONLY);
    if (fd != -1)
    {
        count = lseek(fd, 0, SEEK_END) + 1;
        close(fd);
    }
#endif /*]*/

    return count;
}


static
int readShader(char *fileName, EShaderType shaderType, char *shaderText, int size)
{
    //
    // Reads a shader from the supplied file and returns the shader in the
    // arrays passed in. Returns 1 if successful, 0 if an error occurred.
    // The parameter size is an upper limit of the amount of bytes to read.
    // It is ok for it to be too big.
    //
    FILE *fh;
    char name[100];
    int count;

    strcpy(name, fileName);

    switch (shaderType) 
    {
        case EVertexShader:
            strcat(name, ".vert");
            break;
        case EFragmentShader:
            strcat(name, ".frag");
            break;
        default:
            printf("ERROR: unknown shader file type\n");
            exit(1);
            break;
    }

    //
    // Open the file
    //
    fh = fopen(name, "r");
    if (!fh)
        return -1;

    //
    // Get the shader from a file.
    //
    fseek(fh, 0, SEEK_SET);
    count = (int) fread(shaderText, 1, size, fh);
    shaderText[count] = '\0';

    if (ferror(fh))
        count = 0;

    fclose(fh);
    return count;
}


/*public*/
int readShaderSource(char *fileName, GLchar **vertexShader, GLchar **fragmentShader)
{
    int vSize, fSize;

    //
    // Allocate memory to hold the source of our shaders.
    //
    vSize = shaderSize(fileName, EVertexShader);
    fSize = shaderSize(fileName, EFragmentShader);

    if ((vSize == -1) || (fSize == -1))
    {
        printf("Cannot determine size of the shader %s\n", fileName);
        return 0;
    }

    *vertexShader = (GLchar *) malloc(vSize);
    *fragmentShader = (GLchar *) malloc(fSize);

    //
    // Read the source code
    //
    if (!readShader(fileName, EVertexShader, *vertexShader, vSize))
    {
        printf("Cannot read the file %s.vert\n", fileName);
        return 0;
    }

    if (!readShader(fileName, EFragmentShader, *fragmentShader, fSize))
    {
        printf("Cannot read the file %s.frag\n", fileName);
        return 0;
    }

    return 1;
}


/*public*/
int installBrickShaders(const GLchar *brickVertex,
                        const GLchar *brickFragment)
{
    GLuint brickVS, brickFS, brickProg;   // handles to objects
    GLint  vertCompiled, fragCompiled;    // status values
    GLint  linked;

    // Create a vertex shader object and a fragment shader object
	//
    brickVS = glCreateShader(GL_VERTEX_SHADER);
    brickFS = glCreateShader(GL_FRAGMENT_SHADER);

    // Load source code strings into shaders
	//
    glShaderSource(brickVS, 1, &brickVertex, NULL);
    glShaderSource(brickFS, 1, &brickFragment, NULL);

    // Compile the brick vertex shader, and print out
    // the compiler log file.
	//
    glCompileShader(brickVS);
    printOpenGLError();  // Check for OpenGL errors
    glGetShaderiv(brickVS, GL_COMPILE_STATUS, &vertCompiled);
    printShaderInfoLog(brickVS);

    // Compile the brick vertex shader, and print out
    // the compiler log file.
	//
    glCompileShader(brickFS);
    printOpenGLError();  // Check for OpenGL errors
    glGetShaderiv(brickFS, GL_COMPILE_STATUS, &fragCompiled);
    printShaderInfoLog(brickFS);

    if (!vertCompiled || !fragCompiled)
        return 0;

    // Create a program object and attach the two compiled shaders
	//
    brickProg = glCreateProgram();
    glAttachShader(brickProg, brickVS);
    glAttachShader(brickProg, brickFS);

    // Link the program object and print out the info log
	//
    glLinkProgram(brickProg);
    printOpenGLError();			// Check for OpenGL errors
    glGetProgramiv(brickProg, GL_LINK_STATUS, &linked);
    printProgramInfoLog(brickProg);

    if (!linked)
        return 0;

    // Install program object as part of current state
	//
    glUseProgram(brickProg);

    // Set up initial uniform values
	//
	Video              = getUniLoc(brickProg, "Video");				// Video flag
	RenderBrickPattern = getUniLoc(brickProg, "RenderBrickPattern");// Render or not the brick pattern flag
	SwitchShaders      = getUniLoc(brickProg, "SwitchShaders");     // Switch among different shaders

	// The brick geometry parameters that we'll be using are illustrated in Figure 6.1 of OpenGL SL Book.
	//

	// Parameters that control the brick color, 
	// mortar color, 
	// brick-to-brick horizontal distance, 
	// brick-to-brick vertical distance, 
	// brick width fraction (ratio of the width of a brick to the overall horizontal distance between two adjacent bricks), 
	// brick height fraction (ratio of the height of a brick to the overall vertical distance between two adjacent bricks).
	// brick depth fraction  (ratio of the depth of a brick to the overall z distance between two adjacent bricks).
	//
	// Brick size and brick percentage parameters will both be stored in user-defined uniform variables of type vec3. 
	// The horizontal distance between two bricks, including the width of the mortar, will be provided by BrickSize.x. 
	// The vertical distance between two rows of bricks, including the height of the mortar, will be provided by BrickSize.y. 
	// The z distance between two bricks, including the z of the mortar, will be provided by BrickSize.z. 
	// These three values will be given in units of modeling coordinates. The fraction of BrickSize.x represented by the brick 
	// only will be provided by BrickPct.x. The fraction of BrickSize.y represented by the brick only will be provided by 
	// BrickPct.y and same in z. 
	// These three values will be in the range [0,1]. Finally, the brick color and the mortar color will be represented 
	// by the variables BrickColor and MortarColor. 
	//
	// The values of BrickPct.x, BrickPct.y and BrickPct.z can be computed by the application to give a uniform mortar width 
	// in the three directions based on the ratio of column width to row height, or they can be chosen arbitrarily to give a
	// mortar appearance that looks right. 
	//
    glUniform3f(getUniLoc(brickProg, "BrickColor"), 1.0, 0.3, 0.2);
    glUniform3f(getUniLoc(brickProg, "MortarColor"), 0.85, 0.86, 0.84);

	// We play with BrickSize to make the brick uniform across borders :-)
	// For Brick height (no moldar):
	// Num of Bricks * Brick height (no moldar) + (Num of Bricks - 1) * Moldar height = 2 (units in model coord)
	// If Moldar height = Factor of 'Brick height', let say 10%, and we want 12 brick up: 
	// Brick height = 2 / ( 12 + 1.2 - 0.1)
	// Brick height = 0.15267175572519083969465648854962 -> BrickSize.y = Brick height + 0.1 Brick height
	// BrickSize.y  = 0.1679389312977099236641221374045 including a moldar of 10% (BricPct.y = 0.90)
	// 
	// For Brick width (no moldar):
	// Num of Bricks * (Brick width + Moldar width) + Brick depth (the Brick corner of the next wall) = 2 (units in model coord)
	// If Moldar width = Factor of 'Brick width', let say 12%, and we want 6 bricks across:
	// Let's define a Breick depth (z axis) of 0.10, then Brick width = (2 - 0.10) / 6 * (1 + 0.12)
	// Brick width = 0.2827380952380952380952380952381 -> BrickSize.x = Brick width + 0.12 Brick width
	// BrickSize.x = 0.31666666666666666666666666666667 inclusing a moldar of 12% (BricPct.x = 0.88)
	//
	// To be consistent, BrickSize.z and Moldar in z (depth) has to be equal to BrickSize.x and Moldar in x (width).
	// All walls has to look the same...
	//
	// UFFFFFFFFFFFFFFFF...I hope evrything is right!
	//
	glUniform3f(getUniLoc(brickProg, "BrickSize"), 
													0.31666666666666666666666666666667, 
													0.1679389312977099236641221374045, 
													0.31666666666666666666666666666667);
    glUniform3f(getUniLoc(brickProg, "BrickPct"),  0.88, 0.90, 0.88);

	glUniform3f( LightPosition = getUniLoc(brickProg, "LightPosition"), lightX, lightY, lightZ);
	glUniform1iARB(TexLoc = getUniLoc(brickProg, "TexLoc"), 0); // 0 because we bound to texture 0 in the main program

	glUniform3f( getUniLoc(brickProg, "BaseColor"), 1.0, 0.50, 0.00);
	glUniform1f( getUniLoc(brickProg, "MixRatio"), 0.5);
	glUniform1i(SwitchShaders, 1);
	
	return 1;
}
